local gamera, editgrid = require "gamera", require "editgrid"

local drawings = {}

local checker = love.graphics.newImage("checker.png")
checker:setWrap("repeat","repeat")
checker:setFilter("nearest", "nearest")
local checkw, checkh = checker:getDimensions()

local visuals = {
    size            = 8,
    drawScale       = false, 
    color           = {0,0,0,70},
    xColor          = {128,0,0},
    yColor          = {0,0,255},
    subdivisions    = function (size, zoom)
        local value = math.floor(zoom / 4)
        value = (value % 2 == 0 and value ~= 0) and value or value + 1
        if value > size then return size else return value end
    end
}

drawings.new = function (width, height, sx, sy, sw, sh)
    local sx = sx or 0
    local sy = sy or 0
    local sw = sw or (love.graphics.getWidth()  - sx)
    local sh = sh or (love.graphics.getHeight() - sy)

    local self = {
        width   = width,
        height  = height,
        visuals = visuals,
        shadow  = true
    }

    local w, h = (math.ceil(sw/width) + 2) * width, (math.ceil(sh/height) + 2) * height
    local l, t = -(w - width)/2, -(h - height)/2
    self.cam = gamera.new(l, t, w, h)
    self.cam:setPosition(width * .5, height * .5)

    self.cam:setWindow(sx, sy, sw, sh)

    self.quad = love.graphics.newQuad(0, 0, width, height, checkw, checkh)

    return setmetatable(self, {__index = drawings})
end

drawings.visuals = function (self, vis)
    if not vis then
        visuals = self
    else
        self.visuals = vis
    end
end

drawings.resize = function (self, width, height)
    local posX, posY = self:getPosition()
    local sizW, sizH = self:getSize()
    local sx, sy, sw, sh = self:getViewport()
    local angle, zoom = self:getRotation(), self:getZoom()

    local new = drawings.new(width, height, sx, sy, sw, sh)

    new:setZoom(zoom)
    new:rotate(angle)
    new:move((posX - sizW * .5) / sizW, (posY - sizH * .5) / sizH)

    self = new
end

drawings.getSize = function (self)
    return self.width, self.height
end

drawings.setViewport = function (self, x, y, w, h)
    self.cam:setWindow(x, y, w, h)
end

drawings.getViewport = function (self)
    return self.cam:getWindow()
end

drawings.zoom = function (self, mode, factor, x, y)
    local omx, omy

    if x and y then
        omx, omy = self.cam:toWorld(x, y)
    end

    self.cam:setScale((factor or 2) ^ (mode == "out" and -1 or 1) * self.cam:getScale())

    if x and y then
        local cx, cy = self.cam:getPosition()
        local mx, my = self.cam:toWorld(x, y)
        self.cam:setPosition(cx + (omx - mx), cy + (omy - my))
    end
end

drawings.setZoom = function (self, zoom, x, y)
    local omx, omy

    if x and y then
        omx, omy = self.cam:toWorld(x, y)
    end

    self.cam:setScale(zoom)

    if x and y then
        local cx, cy = self.cam:getPosition()
        local mx, my = self.cam:toWorld(x, y)
        self.cam:setPosition(cx + (omx - mx), cy + (omy - my))
    end
end

drawings.getZoom = function (self)
    return self.cam:getScale()
end

drawings.move = function (self, dx, dy)
    local s, c = self.cam.sin, self.cam.cos
    local x, y = self.cam:getPosition()
    local scale = self.cam:getScale()

    local dx, dy = -dx / scale, -dy / scale

    self.cam:setPosition(x + dx * c - dy * s, y + dy * c + dx * s)
end
--Probably should be used because the position is all weird but whatever
drawings.setPosition = function (self, x, y)
    self.cam:setPosition(x, y)
end

drawings.getPosition = function (self)
    return self.cam:getPosition()
end

drawings.rotate = function (self, dangle)
    self.cam:setAngle(self.cam:getAngle() + dangle)
end

drawings.setRotation = function (self, angle)
    self.cam:setAngle(angle)
end

drawings.getRotation = function (self)
    return self.cam:getAngle()
end

drawings.toWorld = function (self, x, y)
    return self.cam:toWorld(x,y)
end

drawings.toScreen = function (self, x, y)
    return self.cam:toScreen(x,y)
end

drawings.draw = function (self, image)
    if drawings.shadow and self.shadow then
    
    end
    self.cam:draw(function (l,t,w,h)
        love.graphics.setColor(255,255,255)
        love.graphics.draw(checker, self.quad, 0, 0)
        if image then
            love.graphics.draw(image, 0, 0)
        end
    end)
    editgrid.draw(self.cam, self.visuals)
end

drawings.reset = function ()
    self.cam:setScale(1)
    self.cam:setAngle(0)
    self.cam:setPosition(self.width * .5, self.height * .5)
end

return drawings